=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed March 2010 by Giovanni Di Trapani

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   kicad_parser.rb
# Version	:	2.0
# Type		:   Sketchup Tool
# Description	:   Ruby procedures for Kicad .brd file parsing
# Usage		:   Used by kicad.rb
# Original Date	:   18 March 2010
# Last change : 	16 November 2011
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'kicad/kicad_class.rb'

class LEVEL_ACTIONS 
	STAY=0
	EXIT=1
end

class Sections

	NO_BOARD=0
	BOARD_IN=1
	BOARD_OUT=2
	GENERAL=3
	SHEET=4
	SETUP=5
	EQUIPOT=6
	MODULE=7
	PAD=8
	SHAPE3D=9
	DRAWSEGMENT=10
	TEXTPCB=11
	MIREPCB=12
	COTATION=13
	TRACK=14
	ZONE=15
	CZONE_OUTLINE=16
	POLYSCORNERS=17
	
	attr_accessor :active, :string
	
	def initialize()
		@active=NO_BOARD
		@transition={NO_BOARD=>BOARD_IN, BOARD_IN=>BOARD_OUT, GENERAL=>BOARD_IN, SHEET=>BOARD_IN, SETUP=>BOARD_IN, EQUIPOT=>BOARD_IN,
					MODULE=>BOARD_IN, PAD=>MODULE, SHAPE3D=>MODULE, DRAWSEGMENT=>BOARD_IN, TEXTPCB=>BOARD_IN, MIREPCB=>BOARD_IN,
					COTATION=>BOARD_IN, TRACK=>BOARD_IN, ZONE=>BOARD_IN, CZONE_OUTLINE=>BOARD_IN,POLYSCORNERS=>CZONE_OUTLINE}
		@string=["PCBNEW-BOARD missing","BOARD","","GENERAL","SHEET","SETUP","EQUIPOT","MODULE","PAD","SHAPE3D","DRAWSEGMENT",
				"TEXTPCB","MIREPCB","COTATION","TRACK","ZONE","CZONE_OUTLINE","POLYSCORNERS"]
	end
	
	def enter_section(section)
		@active=section
	end
	
	def exit_from_section()
		@active=@transition[@active]
	end
	
end

kicad_section=Sections.new

date_expr = [/\d{1,2}\/\d{1,2}\/\d{4,4}(\s+|-)\d{1,2}:\d{1,2}:\d{1,2}(\s+(AM|PM))*/,/(\w+\s+){2,2}\d+\s+\d{1,2}:\d{1,2}:\d{1,2}\s+\d{4,4}(\s+(AM|PM))*/]
version_expr = /\d+/


#procedures used to scan the PCB file

$goto_section = proc { |pattern|
	
	result=LEVEL_ACTIONS::STAY	
	while(line=$brd.gets)
#		print("scanning",line)
		pattern.each_index do |index|
#			print("\t with pattern ",pattern[index],"\n")
			if(!(index%2)) then next
			end
			if(pattern[index]=~line)
				result=pattern[index+1].call(line,pattern[index+2])
				break
			end
		end
		if(result==LEVEL_ACTIONS::EXIT)
			return LEVEL_ACTIONS::STAY
		end
	end
	if(kicad_section.active!=Sections::BOARD_OUT)
		if(kicad_section==Sections::NO_BOARD)
			UI.messagebox("#{kicad_section.string[kicad_section.active]}")
		else
			UI.messagebox("#{kicad_section.string[kicad_section.active]} not closing properly")
		end
		return 0
	end
}

close_section = proc { |line,pattern|
	kicad_section.exit_from_section()
	LEVEL_ACTIONS::EXIT
}


# procedures used to recognize and draw the PCB

board_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::BOARD_IN)
	$pcb=Board.new
#	print("board detected\n")
#	return
	version=line.scan(/Version\s+\d+/)
	$pcb.file_vers=version[0].scan(/\d+/)
	$goto_section.call(pattern)
	LEVEL_ACTIONS::STAY
}

draw_board = proc { |line,pattern|
	$pcb.draw
	close_section.call(line,pattern)
}

# procedures used to catch info from GENERAL section


general_section = proc {|line,pattern|
	kicad_section.enter_section(Sections::GENERAL)
	$goto_section.call(pattern)
	LEVEL_ACTIONS::STAY	
}

# procedures used to catch info from SHEET section

sheet_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::SHEET)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}
 
sheet_title = proc {|line,pattern|
	parameters=line.scan(/".*"/)
	text=parameters[0].delete("\"")
	if(text==nil)
		$pcb.title=""
	else
		$pcb.title=text
	end
	LEVEL_ACTIONS::STAY
}

sheet_date = proc {|line,pattern|
	parameters=line.scan(/".*"/)
	text=parameters[0].delete("\"")
	if(text==nil)
		$pcb.date=""
	else
		$pcb.date=text
	end
	LEVEL_ACTIONS::STAY
}

sheet_revision = proc {|line,pattern|
	parameters=line.scan(/".*"/)
	text=parameters[0].delete("\"")
	if(text==nil)
		$pcb.rev="NO REV"
	else
		$pcb.rev=text
	end
	LEVEL_ACTIONS::STAY
}

# procedures used to catch info from SETUP section

setup_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::SETUP)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

setup_unit = proc { |line,pattern|
	parameters=line.split(/\s+/)
	$pcb.unit=parameters[1].to_f
	LEVEL_ACTIONS::STAY
}

setup_layers = proc{ |line,pattern|
	parameters=line.split(/\s+/)
	$pcb.copper_layers=parameters[1].to_i
	LEVEL_ACTIONS::STAY
}

setup_layer_name = proc { |line,pattern|
	parameters=line.scan(/\[\d+\]/)
	layer_num=parameters[0].scan(/\d+/)
	layer_name=line.split(/\[\d+\]\s+/)
	$pcb.layer_name[layer_num[0].to_i]=layer_name[1]
	LEVEL_ACTIONS::STAY
}

#procedure used to recognize the EQUIPOT section
equipot_section = proc {|line,pattern|
	kicad_section.enter_section(Sections::EQUIPOT)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

# procedures used to get and configure kicad modules

module_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::MODULE)
	k_module=Kicad_Module.new
	parameters=line.split(/^\s*\$MODULE\s+/)
	name=parameters[1].scan(/.*\S/)
	k_module.name=name[0]
	$pcb.k_module<<k_module
	$goto_section.call(pattern)
	LEVEL_ACTIONS::STAY
}

module_position = proc { |line,pattern|
	
	parameters=line.split(/\s+/)
	k_module=$pcb.last_module
	k_module.Xpos=parameters[1].to_i
	k_module.Ypos=-parameters[2].to_i
	k_module.rotation=parameters[3].to_i
	k_module.layer=parameters[4].to_i
	LEVEL_ACTIONS::STAY
}

module_field = proc { |line,pattern|

	parameters=line.split(/\s+/)
	f_num=parameters[0].scan(/\d+/)
	k_module=$pcb.last_module

#	print("fnum=",f_num[0].to_i," ")
	if(k_module.name=="SKP")
		if(f_num[0].to_i==0)
			k_module.reference=Kicad_Field.new
			text_with_quotes=line.scan(/".*"/)
			text=text_with_quotes[0].delete("\"")
			if(text[0]!=nil)
				k_module.reference.Text=text
			else
				k_module.reference.Text=""
			end
		end
	else
		if(f_num[0].to_i==$kicad_settings.reference) then
			k_module.reference=Kicad_Field.new
			k_module.reference.Xoff=parameters[1].to_i
			k_module.reference.Yoff=parameters[2].to_i
			k_module.reference.Xsize=parameters[3].to_i
			k_module.reference.Ysize=parameters[4].to_i
			k_module.reference.Orientation=parameters[5].to_i
			text_with_quotes=line.scan(/".*"/)
			text=text_with_quotes[0].delete("\"")
			if(text[0]!=nil)
				k_module.reference.Text=text
			else
				k_module.reference.Text=""
			end
		end
	end
	
	LEVEL_ACTIONS::STAY
}


module_segment = proc { |line,pattern|
	parameters=line.split(/\s+/)
	k_module=$pcb.last_module
	segment=Drawing.new
	segment.type=Drawing::SEGMENT
	segment.Xstart=parameters[1].to_i
	segment.Ystart=-parameters[2].to_i
	segment.Xend=parameters[3].to_i
	segment.Yend=-parameters[4].to_i
	segment.width=parameters[5].to_i
	segment.layer=parameters[6].to_i
	k_module.drawing<<segment
	LEVEL_ACTIONS::STAY
}

module_circle = proc { |line,pattern|
	parameters=line.split(/\s+/)
	k_module=$pcb.last_module
	circle=Drawing.new
	circle.type=Drawing::CIRCLE
	circle.Xcentre=parameters[1].to_i
	circle.Ycentre=-parameters[2].to_i
	circle.Xend=parameters[3].to_i
	circle.Yend=-parameters[4].to_i
	circle.width=parameters[5].to_i
	circle.layer=parameters[6].to_i
	k_module.drawing<<circle
	LEVEL_ACTIONS::STAY
}

module_arc = proc { |line,pattern|
	parameters=line.split(/\s+/)
	k_module=$pcb.last_module
	arc=Drawing.new
	arc.type=Drawing::ARC
	arc.Xcentre=parameters[1].to_i
	arc.Ycentre=-parameters[2].to_i
	arc.Xend=parameters[3].to_i
	arc.Yend=-parameters[4].to_i
	arc.angle=parameters[5].to_i
	arc.width=parameters[6].to_i
	arc.layer=parameters[7].to_i
	k_module.drawing<<arc
	LEVEL_ACTIONS::STAY
}

# procedures used to get and info on $PAD

pad_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::PAD)
	kicad_module=$pcb.last_module
	kicad_module.pad<<Pad.new
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

pad_shape = proc { |line,pattern|
	kicad_module=$pcb.last_module
	pad=kicad_module.last_pad
	parameters=line.split(/\s+/)
	parameters[1]=parameters[1].scan(/\w+/)
	pad.name= (parameters[1].empty?) ? "?" : parameters[1]
	pad_shape = { "C" => 0, "R" => 1, "O" => 2, "T" => 3 }
	pad.shape=pad_shape[parameters[2].to_s]
	pad.Xsize=parameters[3].to_i
	pad.Ysize=parameters[4].to_i
	pad.Xdelta=parameters[5].to_i
	pad.Ydelta=parameters[6].to_i
	pad.Orientation=parameters[7].to_i
	LEVEL_ACTIONS::STAY
}

pad_attributes = proc { |line,pattern|
	pad_type = { "STD" => 0, "SMD" => 1, "CONN" => 2, "HOLE" => 3, "MECA" => 4 }
	kicad_module=$pcb.last_module
	pad=kicad_module.last_pad
	parameters=line.split(/\s+/)
	pad.type=pad_type[parameters[1].to_s]
#	if(pad.type==Pad::SMD) then pad.drill=nil
#	end
	if(pad.type!=Pad::STD) then pad.drill=nil
	end
	pad.layer_mask=parameters[3].hex #hexadecimal value...to be fixed
	LEVEL_ACTIONS::STAY
}

pad_position = proc { |line,pattern|
	kicad_module=$pcb.last_module
	pad=kicad_module.last_pad
	parameters=line.split(/\s+/)
	pad.Xpos=parameters[1].to_i
	pad.Ypos=-parameters[2].to_i
	LEVEL_ACTIONS::STAY
}

pad_drill = proc { |line,pattern|
	kicad_module=$pcb.last_module
	pad=kicad_module.last_pad
	parameters=line.split(/\s+/)
	drill=Kicad_drill.new
	drill.Xdrill=parameters[1].to_i
	drill.Xoff=parameters[2].to_i
	drill.Yoff=parameters[3].to_i
	drill.type=Kicad_drill::CIRCLE
	pad.drill=drill
	LEVEL_ACTIONS::STAY
}

pad_oblong_drill = proc { |line,pattern|
	kicad_module=$pcb.last_module
	pad=kicad_module.last_pad
	parameters=line.split(/\s+/)
	drill=Kicad_drill.new
	drill.Xdrill=parameters[5].to_i
	drill.Ydrill=parameters[6].to_i
	drill.Xoff=parameters[2].to_i
	drill.Yoff=parameters[3].to_i
	drill.type=Kicad_drill::OBLONG
	pad.drill=drill
	LEVEL_ACTIONS::STAY
}

# procedures used to get and info on SHAPE3D

shape3d_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::SHAPE3D)
	kicad_module=$pcb.last_module
	kicad_module.shape3d=Shape3D.new
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

shape3d_filename = proc { |line,pattern|

	filename=line.scan(/[\w|-]+\.wrl/)
	basename=filename.last.split(/\.wrl/)
	kicad_module=$pcb.last_module
	shape3d=kicad_module.shape3d
	shape3d.filename=basename[0]
	LEVEL_ACTIONS::STAY	
}


# procedures used to get and info on $DRAWSEGMENT

drawsegment_section = proc {|line,pattern|
	kicad_section.enter_section(Sections::DRAWSEGMENT)
	$pcb.drawsegment<<Drawsegment.new
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

drawsegment_position = proc { |line,pattern|
	
	drawsegment=$pcb.last_drawsegment
	parameters=line.split(/\s+/)
	drawsegment.shape=parameters[1].to_i
	case drawsegment.shape
	when Drawsegment::CIRCLE
		drawsegment.Xcentre=parameters[2].to_i
		drawsegment.Ycentre=-parameters[3].to_i
	when Drawsegment::ARC
		drawsegment.Xcentre=parameters[2].to_i
		drawsegment.Ycentre=-parameters[3].to_i
	when Drawsegment::LINE
		drawsegment.Xstart=parameters[2].to_i
		drawsegment.Ystart=-parameters[3].to_i
	end
	drawsegment.Xend=parameters[4].to_i
	drawsegment.Yend=-parameters[5].to_i
	LEVEL_ACTIONS::STAY
	
}

drawsegment_description = proc { |line,pattern|
	drawsegment=$pcb.last_drawsegment
	parameters=line.split(/\s+/)	
	drawsegment.layer=parameters[1].to_i
	drawsegment.angle=parameters[3].to_i
	LEVEL_ACTIONS::STAY
}

#import settings for TEXTPCB section

textpcb_section = proc {|line,pattern|
	kicad_section.enter_section(Sections::TEXTPCB)
	$pcb.text<<Kicad_text.new
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

text_content = proc {|line,pattern|
	parameters=line.scan(/".*"/)
	text=parameters[0].delete("\"")
	last_text=$pcb.text.last
	last_text.text=text
	LEVEL_ACTIONS::STAY
}

text_position = proc {|line,pattern|

	parameters=line.split(/\s+/)
	last_text=$pcb.text.last
	last_text.Xstart=parameters[1].to_i
	last_text.Ystart=parameters[2].to_i
	last_text.Xsize=parameters[3].to_i
	last_text.Ysize=parameters[4].to_i
	last_text.rotation=parameters[6].to_i
	LEVEL_ACTIONS::STAY
}

text_description = proc {|line,pattern|
	parameters=line.split(/\s+/)
	last_text=$pcb.text.last
	last_text.layer=parameters[1].to_i
	last_text.normal=parameters[2].to_i
	LEVEL_ACTIONS::STAY
}


mirepcb_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::MIREPCB)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

cotation_section = proc {|line,pattern|
	kicad_section.enter_section(Sections::COTATION)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

track_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::TRACK)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

zone_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::ZONE)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}

polyscorners_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::POLYSCORNERS)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}


czone_outline_section = proc { |line,pattern|
	kicad_section.enter_section(Sections::CZONE_OUTLINE)
	$goto_section.call(pattern) 
	LEVEL_ACTIONS::STAY
}



null = proc { |line, pattern|
	LEVEL_ACTIONS::STAY
}


$pattern = [

#	/PCBNEW-BOARD\s+Version\s+#{version_expr}\s+date\s+(#{date_expr[0]}|#{date_expr[1]})\s*$/, board_section,
	/PCBNEW-BOARD\s+Version\s+#{version_expr}\s+date\s+.*$/, board_section,
	[
		/^\s*\$GENERAL\s*$/, general_section,
		[	/^\s*Ly\s+(\d|[A-F]){8,8}\s*$/, null,
			/^\s*Links\s+\d+\s*$/, null,
			/^\s*NoConn\s+\d+\s*$/, null,
			/^\s*Di(\s+\d+){4,4}\s*$/, null,
			/^\s*Ndraw\s+\d+\s*$/, null,
			/^\s*Ntrack\s+\d+\s*$/, null,
			/^\s*Nzone\s+\d+\s*$/, null,
			/^\s*Nmodule\s+\d+\s*$/, null,
			/^\s*Nnets\s+\d+\s*$/, null,
			/^\s*\$EndGENERAL\s*$/, close_section
		],
		/^\s*\$SHEETDESCR\s*$/, sheet_section,
		[	/^\s*Sheet\s+\w+\s+\d+\s+\d+\s*$/, null,
			/^\s*Title\s+"(\s|\w)*"\s*$/, sheet_title,
			/^\s*Date\s+"(\s|\w)*"\s*$/, sheet_date,
			/^\s*Rev\s+"(\s|\w)*"\s*$/, sheet_revision,
			/^\s*Comp\s+"(\s|\w)*"\s*$/, null,
			/^\s*Comment\d+\s+"(\s|\w)*"\s*$/, null,
			/^\s*\$EndSHEETDESCR\s*$/, close_section
		],
		/^\s*\$SETUP\s*$/, setup_section,
		[	/^\s*InternalUnit\s+(\d|\.)+\s+INCH\s*$/, setup_unit, #check if INCH can change
			/^\s*Layers\s+\d+\s*$/, setup_layers,
			/^\s*Layer\[\d+\]\s+.+\s*$/, setup_layer_name,
			/^\s*TrackWidthHistory\s+\d+\s*$/, null,
			/^\s*TrackClearence\s+\d+\s*$/, null,
			/^\s*ZoneClearence\s+\d+\s*$/, null,
			/^\s*DrawSegmWidth\s+\d+\s*$/, null,
			/^\s*ViaSize\s+\d+\s*$/, null,
			/^\s*ViaDrill\s+\d+\s*$/, null,
			/^\s*ViaSizeHistory\s+\d+\s*$/, null,
			/^\s*TextPcbWidth\s+\d+\s*$/, null,
			/^\s*TextPcbSize(\s+\d+){2,2}\s*$/, null,
			/^\s*EdgeModWidth\s+\d+\s*$/, null,
			/^\s*TextModSize(\s+\d+){2,2}\s*$/, null,
			/^\s*TextModWidth\s+\d+\s*$/, null,
			/^\s*PadSize(\s+\d+){2,2}\s*$/, null,
			/^\s*PadDrill\s+\d+\s*$/, null,
			/^\s*AuxiliaryAxisOrg(\s+\d+){2,2}\s*$/, null,
			/^\s*\$EndSETUP\s*$/, close_section
		],			
		/^\s*\$EQUIPOT\s*$/, equipot_section,
		[	/^\s*\Na\s+\d+\s+".+"\s*$/, null,
			/^\s*St\s+~\s*$/, null,
			/^\s*\$EndEQUIPOT\s*$/, close_section
		],	
		/^\s*\$MODULE\s+.+\s*$/, module_section,
		[	/^\s*Po\s+\d+\s+\d+\s+\d+\s+\d+\s+(\d|[A-F])+\s+(\d|[A-F])+\s+(~|F)(~|P)\s*$/, module_position,
			/^\s*Li\s+(\w|-)+\s*$/, null,
			/^\s*Cd.+\s*$/, null,
			/^\s*Kw\s+(\w+\s+)+\s*$/, null,
			/^\s*Sc\s+(\d|[A-F])+\s*$/, null,
			/^\s*Op(\s+\d+){3,3}\s*$/, null,
			/^\s*T\d+(\s+-?\d+){6,6}\s+(N|M)\s+(I|V)\s+\d+\s+N?\s*".*"\s*$/, module_field,
			/^\s*DS(\s+-?\d+){4,4}(\s+\d+){2,2}\s*$/, module_segment,
			/^\s*DC(\s+-?\d+){4,4}(\s+\d+){2,2}\s*$/, module_circle,
			/^\s*DA(\s+-?\d+){4,4}(\s+\d+){3,3}\s*$/, module_arc,
			/^\s*\$SHAPE3D\s*$/, shape3d_section,
			[	/^\s*Na\s+".+"\s*$/, shape3d_filename,
				/^\s*Sc(\s+(\d|\.)+){3,3}\s*$/, null,
				/^\s*Of(\s+(\d|\.)+){3,3}\s*$/, null,
				/^\s*Ro(\s+(\d|\.)+){3,3}\s*$/, null,
				/^\s*\$EndSHAPE3D\s*$/, close_section
			],
			/^\s*\$PAD\s*$/, pad_section,
			[	/^\s*Sh\s+".*"\s+(C|R|O|T)(\s+\d+){4,4}(\s+-?\d+)\s*$/, pad_shape,
				/^\s*Dr\s+\d+(\s+-?\d+){2,2}\s*$/, pad_drill,
				/^\s*Dr\s+\d+(\s+-?\d+){2,2}\s+O(\s+\d+){2,2}\s*$/, pad_oblong_drill,
				/^\s*At\s+(STD|SMD|CONN|HOLE|MECA)\s+N\s+(\d|[A-F])+\s*$/, pad_attributes,
				/^\s*Ne\s+\d+\s+".*"\s*$/, null,
				/^\s*Po(\s+-?\d+){2,2}\s*$/, pad_position,
				/^\s*\$EndPAD\s*$/, close_section
			],
			/^\s*\$EndMODULE\s+.+\s*$/, close_section
		],

		/^\s*\$DRAWSEGMENT\s*$/, drawsegment_section,
		[
			/^\s*Po\s+\d+(\s+-?\d+){4,4}\s+\d+\s*$/, drawsegment_position,
			/^\s*De(\s+\d+){3,3}\s+(\d|[A-F])+\s+\d+\s*$/, drawsegment_description,
			/^\s*\$EndDRAWSEGMENT\s*$/, close_section
		],

		/^\s*\$TEXTPCB\s*$/, textpcb_section,
		[
			/^\s*Te\s+".+"\s*$/, text_content,
			/^\s*Po(\s+-?\d+){2,2}(\s+\d+){4,4}\s*$/, text_position,
			/^\s*De(\s+\d+){2,2}\s+(\d|[A-F])+\s+\w+\s*$/, text_description,
			/^\s*\$EndTEXTPCB\s*$/, close_section
		],
			
		/^\s*\$MIREPCB\s*$/, mirepcb_section,
		[
			/^\s*Po\s+\d+(\s+-?\d+){2,2}(\s+\d+){2,2}(\d|[A-F])+\s*$/, null,
			/^\s*\$EndMIREPCB\s*$/, close_section
		],

		/^\s*\$COTATION\s*$/, cotation_section,
		[
			/^\s*Ge(\s+\d+){2,2}\s+(\d|[A-F])+\s*$/, null,
			/^\s*Po(\s+-?\d+){2,2}(\s+\d+){5,5}\s*$/, null,
			/^\s*\$(E|e)ndCOTATION\s*$/, close_section
		],

		
		/^\s*\$TRACK\s*$/, track_section,
		[	/^\s*Po\s+\d+(\s+-?\d+){4,4}\s+\d+\s+-?\d+\s*$/, null,
			/^\s*De(\s+\d+){3,3}\s+(\d|[A-F])+\s+\d+\s*$/, null,
			/^\s*\$EndTRACK\s*$/, close_section
		],

		/^\s*\$ZONE\s*$/, zone_section,
		[	/^\s*Po\s+\d+(\s+-?\d+){4,4}\s+\d+\s+-?\d+\s*$/, null,
			/^\s*De(\s+\d+){3,3}\s+(\d|[A-F])+\s+\d+\s*$/, null,
			/^\s*\$EndZONE\s*$/, close_section
		],
		
		/^\s*\$CZONE_OUTLINE\s*$/, czone_outline_section,
		[	/^\s*Zinfo\s+(\d+|[A-F])+\s+\d+\s+".+"\s*$/, null,
			/^\s*ZLayer\s+\d+\s*$/, null,
			/^\s*ZAux\s+\d+\s+(N|E|F)\s*$/, null,
			/^\s*ZClearence\s+\d+\s+(I|T|X)\s*$/, null,
			/^\s*ZMinThickness\s+\d+\s*$/, null,
			/^\s*ZOptions(\s+\d+){2,2}\s+F(\s+\d+){2,2}\s*$/, null,
			/^\s*ZCorner(\s+\d+){3,3}\s*$/, null,
			/^\s*\$POLYSCORNERS\s*$/, polyscorners_section,
			[
				/^\s*(-?\d+\s+){2,2}(\d+\s*){2,2}$/, null,
				/^\s*\$endPOLYSCORNERS\s*$/, close_section
			],	
			/^\s*\$endCZONE_OUTLINE\s*$/, close_section
		],
		/^\s*\$EndBOARD/, draw_board
	]	
]

